version: 1

# Rules to match for a machine to qualify
target:
  id: "{{ machine_id }}"

timeouts:
  first_console_activity:  # This limits the time it can take to receive the first console log
    minutes: 2
    retries: 0
  overall:           # Maximum time the job can take, not overrideable by the "continue" deployment
    minutes: 3
    retries: 0
    # no retries possible here

console_patterns:
    session_end:
        regex: "^.*It's now safe to turn off your computer\r$"
    job_success:
        regex: "^.*Execution is over, pipeline status: 0\r$"
    job_warn:
        regex: "^The configured console is not connected to SALAD$"

# Environment to deploy
deployment:
  # Initial boot
  start:
    kernel:
      cmdline:
        - b2c.container="-ti --tls-verify=false docker://{{ fdo_proxy_registry }}/gfx-ci/ci-tron/machine-registration:latest check"
        - b2c.ntp_peer="ci-gateway" b2c.pipefail b2c.cache_device=auto b2c.poweroff_delay=15
        - console={{ local_tty_device }},115200 earlyprintk=vga,keep SALAD.machine_id={{ machine_id }}
        - loglevel=6 nomodeset
