version: 1

# Rules to match for a machine to qualify
target:
  id: '{{ machine_id }}'

timeouts:
  overall:           # Maximum time the job can take, not overrideable by the "continue" deployment
    minutes: 5
    retries: 0
    # no retries possible here

console_patterns:
    session_end:
        regex: "^MaRS: Registration .*$"
    job_success:
        regex: "^MaRS: Registration complete$"

# Environment to deploy
deployment:
  # Initial boot
  start:
    kernel:
      cmdline:
        - b2c.container="-ti --tls-verify=false docker://{{ fdo_proxy_registry }}/gfx-ci/ci-tron/machine-registration:latest register"
        - b2c.ntp_peer="ci-gateway" b2c.cache_device=auto loglevel=6 b2c.poweroff_delay=15 nomodeset
        - console=tty0 console=ttynull # Use the dummy console if possible, but default to a FB-based one if missing
